#include "test_object.h"
#include "core/render.h"
#include "utils/debug.h"
#include "core/input.h"
#include "raylib.h"

START_REFLECT(TestObject);
REFLECT_TYPECLASS(TestObject, Drop);
REFLECT_TYPECLASS(TestObject, SceneNodeEntity);
REFLECT_TYPECLASS(TestObject, Renderable);
END_REFLECT(TestObject);

impl_Drop_for(TestObject,
    DestroyTestObject
)

impl_Renderable_for(TestObject,
    TestObjectDraw
)

impl_SceneNodeEntity_for(TestObject,
    TestObjectEnterTree,
    TestObjectExitTree,
    TestObjectTick
)

SceneNode *CreateTestObject() {
    TestObject *self = new(TestObject);
    self->rotation = self->fly_input = Vector2Zero();
    self->stopped = false;
    return CreateSceneNode(TestObject_as_SceneNodeEntity(self));
}

void DestroyTestObject(TestObject *self) {
    free(self);
}

void TestObjectEnterTree(TestObject *self) {
    self->transform = TC_CAST(self->node->parent->entity, Transformable);
    AddRenderable(TestObject_as_Renderable(self));
    AddListener("pitch_up", ButtonInputListener(self, TestObjectUpInput));
    AddListener("pitch_down", ButtonInputListener(self, TestObjectDownInput));
    AddListener("roll_left", ButtonInputListener(self, TestObjectLeftInput));
    AddListener("roll_right", ButtonInputListener(self, TestObjectRightInput));
    AddListener("stop", ButtonInputListener(self, TestObjectStopInput));
}

void TestObjectExitTree(TestObject *self) {
    RemoveRenderable(TestObject_as_Renderable(self));
}

void TestObjectDraw(TestObject *self) {
}

void TestObjectTick(TestObject *self, double delta) {
    if(self->stopped) return;
    Vector2 diff = Vector2Subtract(self->fly_input, self->rotation);
    float const length = Vector2Length(diff);
    if(length != 0.f)
        self->rotation = Vector2Add(self->rotation, Vector2Scale(diff, 1.0f/length * fminf(4.f * delta, length)));
    Transform global_transform = self->transform.tc->get_global_transform(self->transform.data);
    Matrix global_matrix = TransformGetMatrix(global_transform);
    global_transform.translation = Vector3Add(global_transform.translation, Vector3Scale(MATRIX_FORWARD(global_matrix), 20.f * delta));
    global_transform.rotation = QuaternionMultiply(QuaternionFromAxisAngle(MATRIX_FORWARD(global_matrix), self->rotation.x * 4.f * delta), global_transform.rotation);
    global_transform.rotation = QuaternionMultiply(QuaternionFromAxisAngle(MATRIX_RIGHT(global_matrix), self->rotation.y * 4.f * delta), global_transform.rotation);
    self->transform.tc->set_global_transform(self->transform.data, global_transform);
}


void TestObjectLeftInput(TestObject *self, bool value) {
    self->fly_input.x += value ? -1 : +1;
}

void TestObjectRightInput(TestObject *self, bool value) {
    self->fly_input.x += value ? +1 : -1;
}

void TestObjectUpInput(TestObject *self, bool value) {
    self->fly_input.y += value ? -1 : +1;
}

void TestObjectDownInput(TestObject *self, bool value) {
    self->fly_input.y += value ? +1 : -1;
}

void TestObjectStopInput(TestObject *self, bool value) {
    self->stopped = value;
}
